@@ -1,8 +1,24 @@
+# Copyright 2013-2014 DataStax, Inc.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+# http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 """
 This module houses the main classes you will interact with,
 :class:`.Cluster` and :class:`.Session`.
 """
+from __future__ import absolute_import
 
+from collections import defaultdict
 from concurrent.futures import ThreadPoolExecutor
 import logging
 import socket
@@ -31,21 +47,24 @@ from cassandra.decoder import (QueryMessage, ResultMessage,
                                PreparedQueryNotFound,
                                IsBootstrappingErrorMessage, named_tuple_factory,
                                dict_factory)
-from cassandra.metadata import Metadata
-from cassandra.metrics import Metrics
+from cassandra.metadata import Metadata, protect_name
 from cassandra.policies import (RoundRobinPolicy, SimpleConvictionPolicy,
                                 ExponentialReconnectionPolicy, HostDistance,
                                 RetryPolicy)
 from cassandra.pool import (_ReconnectionHandler, _HostReconnectionHandler,
-                            HostConnectionPool)
+                            HostConnectionPool, NoConnectionsAvailable)
 from cassandra.query import (SimpleStatement, PreparedStatement, BoundStatement,
                              bind_params, QueryTrace, Statement)
 
-# libev is all around faster, so we want to try and default to using that when we can
-try:
-    from cassandra.io.libevreactor import LibevConnection as DefaultConnection
-except ImportError:
-    from cassandra.io.asyncorereactor import AsyncoreConnection as DefaultConnection  # NOQA
+# default to gevent when we are monkey patched, otherwise if libev is available, use that as the
+# default because it's faster than asyncore
+if 'gevent.monkey' in sys.modules:
+    from cassandra.io.geventreactor import GeventConnection as DefaultConnection
+else:
+    try:
+        from cassandra.io.libevreactor import LibevConnection as DefaultConnection  # NOQA
+    except ImportError:
+        from cassandra.io.asyncorereactor import AsyncoreConnection as DefaultConnection  # NOQA
 
 # Forces load of utf8 encoding module to avoid deadlock that occurs
 # if code that is being imported tries to import the module in a seperate
@@ -276,13 +295,6 @@ class Cluster(object):
         Any of the mutable Cluster attributes may be set as keyword arguments
         to the constructor.
         """
-        if 'gevent.monkey' in sys.modules:
-            raise Exception(
-                "gevent monkey-patching detected. This driver does not currently "
-                "support gevent, and monkey patching will break the driver "
-                "completely. You can track progress towards adding gevent "
-                "support here: https://datastax-oss.atlassian.net/browse/PYTHON-7.")
-
         self.contact_points = contact_points
         self.port = port
         self.compression = compression
@@ -364,6 +376,7 @@ class Cluster(object):
         self._lock = RLock()
 
         if self.metrics_enabled:
+            from cassandra.metrics import Metrics
             self.metrics = Metrics(weakref.proxy(self))
 
         self.control_connection = ControlConnection(
@@ -464,20 +477,21 @@ class Cluster(object):
 
                 self.load_balancing_policy.populate(
                     weakref.proxy(self), self.metadata.all_hosts())
+
+                if self.control_connection:
+                    try:
+                        self.control_connection.connect()
+                        log.debug("Control connection created")
+                    except Exception:
+                        log.exception("Control connection failed to connect, "
+                                      "shutting down Cluster:")
+                        self.shutdown()
+                        raise
+
+                self.load_balancing_policy.check_supported()
+
                 self._is_setup = True
 
-            if self.control_connection:
-                try:
-                    self.control_connection.connect()
-                    log.debug("Control connection created")
-                except Exception:
-                    log.exception("Control connection failed to connect, "
-                                  "shutting down Cluster:")
-                    self.shutdown()
-                    raise
-
-            self.load_balancing_policy.check_supported()
-
         session = self._new_session()
         if keyspace:
             session.set_keyspace(keyspace)
@@ -758,13 +772,13 @@ class Cluster(object):
             self.on_down(host, is_host_addition, force_if_down=True)
         return is_down
 
-    def add_host(self, address, signal):
+    def add_host(self, address, datacenter=None, rack=None, signal=True):
         """
         Called when adding initial contact points and when the control
         connection subsequently discovers a new node.  Intended for internal
         use only.
         """
-        new_host = self.metadata.add_host(address)
+        new_host = self.metadata.add_host(address, datacenter, rack)
         if new_host and signal:
             log.info("New Cassandra host %s added", address)
             self.on_add(new_host)
@@ -1252,7 +1266,7 @@ class Session(object):
         Set the default keyspace for all queries made through this Session.
         This operation blocks until complete.
         """
-        self.execute('USE "%s"' % (keyspace,))
+        self.execute('USE %s' % (protect_name(keyspace),))
 
     def _set_keyspace_for_all_pools(self, keyspace, callback):
         """
@@ -1323,8 +1337,8 @@ class ControlConnection(object):
     _SELECT_COLUMN_FAMILIES = "SELECT * FROM system.schema_columnfamilies"
     _SELECT_COLUMNS = "SELECT * FROM system.schema_columns"
 
-    _SELECT_PEERS = "SELECT peer, data_center, rack, tokens, rpc_address FROM system.peers"
-    _SELECT_LOCAL = "SELECT cluster_name, data_center, rack, tokens, partitioner FROM system.local WHERE key='local'"
+    _SELECT_PEERS = "SELECT peer, data_center, rack, tokens, rpc_address, schema_version FROM system.peers"
+    _SELECT_LOCAL = "SELECT cluster_name, data_center, rack, tokens, partitioner, schema_version FROM system.local WHERE key='local'"
 
     _SELECT_SCHEMA_PEERS = "SELECT rpc_address, schema_version FROM system.peers"
     _SELECT_SCHEMA_LOCAL = "SELECT schema_version FROM system.local WHERE key='local'"
@@ -1407,8 +1421,13 @@ class ControlConnection(object):
                 "SCHEMA_CHANGE": self._handle_schema_change
             })
 
-            self._refresh_node_list_and_token_map(connection)
-            self._refresh_schema(connection)
+            peers_query = QueryMessage(query=self._SELECT_PEERS, consistency_level=ConsistencyLevel.ONE)
+            local_query = QueryMessage(query=self._SELECT_LOCAL, consistency_level=ConsistencyLevel.ONE)
+            shared_results = connection.wait_for_responses(
+                peers_query, local_query, timeout=self._timeout)
+
+            self._refresh_node_list_and_token_map(connection, preloaded_results=shared_results)
+            self._refresh_schema(connection, preloaded_results=shared_results)
         except Exception:
             connection.close()
             raise
@@ -1419,7 +1438,7 @@ class ControlConnection(object):
         if self._is_shutdown:
             return
 
-        self._cluster.executor.submit(self._reconnect)
+        self._submit(self._reconnect)
 
     def _reconnect(self):
         log.debug("[control connection] Attempting to reconnect")
@@ -1458,6 +1477,14 @@ class ControlConnection(object):
             self._reconnection_handler = new_handler
             return old
 
+    def _submit(self, *args, **kwargs):
+        try:
+            if not self._cluster._is_shutdown:
+                return self._cluster.executor.submit(*args, **kwargs)
+        except ReferenceError:
+            pass
+        return None
+
     def shutdown(self):
         with self._lock:
             if self._is_shutdown:
@@ -1476,12 +1503,17 @@ class ControlConnection(object):
         try:
             if self._connection:
                 self._refresh_schema(self._connection, keyspace, table)
+        except ReferenceError:
+            pass  # our weak reference to the Cluster is no good
         except Exception:
             log.debug("[control connection] Error refreshing schema", exc_info=True)
             self._signal_error()
 
-    def _refresh_schema(self, connection, keyspace=None, table=None):
-        self.wait_for_schema_agreement(connection)
+    def _refresh_schema(self, connection, keyspace=None, table=None, preloaded_results=None):
+        if self._cluster._is_shutdown:
+            return
+
+        self.wait_for_schema_agreement(connection, preloaded_results=preloaded_results)
 
         where_clause = ""
         if keyspace:
@@ -1522,17 +1554,25 @@ class ControlConnection(object):
         try:
             if self._connection:
                 self._refresh_node_list_and_token_map(self._connection)
+        except ReferenceError:
+            pass  # our weak reference to the Cluster is no good
         except Exception:
             log.debug("[control connection] Error refreshing node list and token map", exc_info=True)
             self._signal_error()
 
-    def _refresh_node_list_and_token_map(self, connection):
-        log.debug("[control connection] Refreshing node list and token map")
-        cl = ConsistencyLevel.ONE
-        peers_query = QueryMessage(query=self._SELECT_PEERS, consistency_level=cl)
-        local_query = QueryMessage(query=self._SELECT_LOCAL, consistency_level=cl)
-        peers_result, local_result = connection.wait_for_responses(
-            peers_query, local_query, timeout=self._timeout)
+    def _refresh_node_list_and_token_map(self, connection, preloaded_results=None):
+        if preloaded_results:
+            log.debug("[control connection] Refreshing node list and token map using preloaded results")
+            peers_result = preloaded_results[0]
+            local_result = preloaded_results[1]
+        else:
+            log.debug("[control connection] Refreshing node list and token map")
+            cl = ConsistencyLevel.ONE
+            peers_query = QueryMessage(query=self._SELECT_PEERS, consistency_level=cl)
+            local_query = QueryMessage(query=self._SELECT_LOCAL, consistency_level=cl)
+            peers_result, local_result = connection.wait_for_responses(
+                peers_query, local_query, timeout=self._timeout)
+
         peers_result = dict_factory(*peers_result.results)
 
         partitioner = None
@@ -1546,7 +1586,9 @@ class ControlConnection(object):
 
             host = self._cluster.metadata.get_host(connection.host)
             if host:
-                host.set_location_info(local_row["data_center"], local_row["rack"])
+                datacenter = local_row.get("data_center")
+                rack = local_row.get("rack")
+                self._update_location_info(host, datacenter, rack)
 
             partitioner = local_row.get("partitioner")
             tokens = local_row.get("tokens")
@@ -1564,10 +1606,13 @@ class ControlConnection(object):
             found_hosts.add(addr)
 
             host = self._cluster.metadata.get_host(addr)
+            datacenter = row.get("data_center")
+            rack = row.get("rack")
             if host is None:
                 log.debug("[control connection] Found new host to connect to: %s", addr)
-                host = self._cluster.add_host(addr, signal=True)
-            host.set_location_info(row.get("data_center"), row.get("rack"))
+                host = self._cluster.add_host(addr, datacenter, rack, signal=True)
+            else:
+                self._update_location_info(host, datacenter, rack)
 
             tokens = row.get("tokens")
             if partitioner and tokens:
@@ -1584,11 +1629,22 @@ class ControlConnection(object):
             log.debug("[control connection] Fetched ring info, rebuilding metadata")
             self._cluster.metadata.rebuild_token_map(partitioner, token_map)
 
+    def _update_location_info(self, host, datacenter, rack):
+        if host.datacenter == datacenter and host.rack == rack:
+            return
+
+        # If the dc/rack information changes, we need to update the load balancing policy.
+        # For that, we remove and re-add the node against the policy. Not the most elegant, and assumes
+        # that the policy will update correctly, but in practice this should work.
+        self._cluster.load_balancing_policy.on_down(host)
+        host.set_location_info(datacenter, rack)
+        self._cluster.load_balancing_policy.on_up(host)
+
     def _handle_topology_change(self, event):
         change_type = event["change_type"]
         addr, port = event["address"]
         if change_type == "NEW_NODE":
-            self._cluster.scheduler.schedule(10, self._cluster.add_host, addr, signal=True)
+            self._cluster.scheduler.schedule(10, self.refresh_node_list_and_token_map)
         elif change_type == "REMOVED_NODE":
             host = self._cluster.metadata.get_host(addr)
             self._cluster.scheduler.schedule(0, self._cluster.remove_host, host)
@@ -1602,7 +1658,7 @@ class ControlConnection(object):
         if change_type == "UP":
             if host is None:
                 # this is the first time we've seen the node
-                self._cluster.scheduler.schedule(1, self._cluster.add_host, addr, signal=True)
+                self._cluster.scheduler.schedule(1, self.refresh_node_list_and_token_map)
             else:
                 # this will be run by the scheduler
                 self._cluster.scheduler.schedule(1, self._cluster.on_up, host)
@@ -1620,11 +1676,11 @@ class ControlConnection(object):
         table = event['table'] or None
         if event['change_type'] in ("CREATED", "DROPPED"):
             keyspace = keyspace if table else None
-            self._cluster.executor.submit(self.refresh_schema, keyspace)
+            self._submit(self.refresh_schema, keyspace)
         elif event['change_type'] == "UPDATED":
-            self._cluster.executor.submit(self.refresh_schema, keyspace, table)
+            self._submit(self.refresh_schema, keyspace, table)
 
-    def wait_for_schema_agreement(self, connection=None):
+    def wait_for_schema_agreement(self, connection=None, preloaded_results=None):
         # Each schema change typically generates two schema refreshes, one
         # from the response type and one from the pushed notification. Holding
         # a lock is just a simple way to cut down on the number of schema queries
@@ -1633,14 +1689,24 @@ class ControlConnection(object):
             if self._is_shutdown:
                 return
 
-            log.debug("[control connection] Waiting for schema agreement")
             if not connection:
                 connection = self._connection
 
+            if preloaded_results:
+                log.debug("[control connection] Attempting to use preloaded results for schema agreement")
+
+                peers_result = preloaded_results[0]
+                local_result = preloaded_results[1]
+                schema_mismatches = self._get_schema_mismatches(peers_result, local_result, connection.host)
+                if schema_mismatches is None:
+                    return True
+
+            log.debug("[control connection] Waiting for schema agreement")
             start = self._time.time()
             elapsed = 0
             cl = ConsistencyLevel.ONE
             total_timeout = self._cluster.max_schema_agreement_wait
+            schema_mismatches = None
             while elapsed < total_timeout:
                 peers_query = QueryMessage(query=self._SELECT_SCHEMA_PEERS, consistency_level=cl)
                 local_query = QueryMessage(query=self._SELECT_SCHEMA_LOCAL, consistency_level=cl)
@@ -1654,36 +1720,45 @@ class ControlConnection(object):
                     elapsed = self._time.time() - start
                     continue
 
-                peers_result = dict_factory(*peers_result.results)
-
-                versions = set()
-                if local_result.results:
-                    local_row = dict_factory(*local_result.results)[0]
-                    if local_row.get("schema_version"):
-                        versions.add(local_row.get("schema_version"))
-
-                for row in peers_result:
-                    if not row.get("rpc_address") or not row.get("schema_version"):
-                        continue
-
-                    rpc = row.get("rpc_address")
-                    if rpc == "0.0.0.0":  # TODO ipv6 check
-                        rpc = row.get("peer")
-
-                    peer = self._cluster.metadata.get_host(rpc)
-                    if peer and peer.is_up:
-                        versions.add(row.get("schema_version"))
-
-                if len(versions) == 1:
-                    log.debug("[control connection] Schemas match")
+                schema_mismatches = self._get_schema_mismatches(peers_result, local_result, connection.host)
+                if schema_mismatches is None:
                     return True
 
                 log.debug("[control connection] Schemas mismatched, trying again")
                 self._time.sleep(0.2)
                 elapsed = self._time.time() - start
 
+            log.warn("Node %s is reporting a schema disagreement: %s",
+                     connection.host, schema_mismatches)
             return False
 
+    def _get_schema_mismatches(self, peers_result, local_result, local_address):
+        peers_result = dict_factory(*peers_result.results)
+
+        versions = defaultdict(set)
+        if local_result.results:
+            local_row = dict_factory(*local_result.results)[0]
+            if local_row.get("schema_version"):
+                versions[local_row.get("schema_version")].add(local_address)
+
+        for row in peers_result:
+            if not row.get("rpc_address") or not row.get("schema_version"):
+                continue
+
+            rpc = row.get("rpc_address")
+            if rpc == "0.0.0.0":  # TODO ipv6 check
+                rpc = row.get("peer")
+
+            peer = self._cluster.metadata.get_host(rpc)
+            if peer and peer.is_up:
+                versions[row.get("schema_version")].add(rpc)
+
+        if len(versions) == 1:
+            log.debug("[control connection] Schemas match")
+            return None
+
+        return dict((version, list(nodes)) for version, nodes in versions.iteritems())
+
     def _signal_error(self):
         # try just signaling the cluster, as this will trigger a reconnect
         # as part of marking the host down
@@ -1840,6 +1915,7 @@ class ResponseFuture(object):
         self.default_timeout = default_timeout
         self._metrics = metrics
         self.prepared_statement = prepared_statement
+        self._callback_lock = Lock()
         if metrics is not None:
             self._start_time = time.time()
 
@@ -1891,9 +1967,15 @@ class ResponseFuture(object):
             # TODO get connectTimeout from cluster settings
             connection = pool.borrow_connection(timeout=2.0)
             request_id = connection.send_msg(message, cb=cb)
+        except NoConnectionsAvailable as exc:
+            log.debug("All connections for host %s are at capacity, moving to the next host", host)
+            self._errors[host] = exc
+            return None
         except Exception as exc:
             log.debug("Error querying host %s", host, exc_info=True)
             self._errors[host] = exc
+            if self._metrics is not None:
+                self._metrics.on_connection_error()
             if connection:
                 pool.return_connection(connection)
             return None
@@ -2110,7 +2192,8 @@ class ResponseFuture(object):
                 del self.session  # clear reference cycles
             except AttributeError:
                 pass
-        self._final_result = response
+        with self._callback_lock:
+            self._final_result = response
         self._event.set()
         if self._callback:
             fn, args, kwargs = self._callback
@@ -2123,7 +2206,8 @@ class ResponseFuture(object):
             del self.session  # clear reference cycles
         except AttributeError:
             pass
-        self._final_exception = response
+        with self._callback_lock:
+            self._final_exception = response
         self._event.set()
         if self._errback:
             fn, args, kwargs = self._errback
@@ -2243,10 +2327,14 @@ class ResponseFuture(object):
             >>> future.add_callback(handle_results, time.time(), should_log=True)
 
         """
-        if self._final_result is not _NOT_SET:
+        run_now = False
+        with self._callback_lock:
+            if self._final_result is not _NOT_SET:
+                run_now = True
+            else:
+                self._callback = (fn, args, kwargs)
+        if run_now:
             fn(self._final_result, *args, **kwargs)
-        else:
-            self._callback = (fn, args, kwargs)
         return self
 
     def add_errback(self, fn, *args, **kwargs):
@@ -2255,10 +2343,14 @@ class ResponseFuture(object):
         An Exception instance will be passed as the first positional argument
         to `fn`.
         """
-        if self._final_exception:
+        run_now = False
+        with self._callback_lock:
+            if self._final_exception:
+                run_now = True
+            else:
+                self._errback = (fn, args, kwargs)
+        if run_now:
             fn(self._final_exception, *args, **kwargs)
-        else:
-            self._errback = (fn, args, kwargs)
         return self
 
     def add_callbacks(self, callback, errback,
